import operator

import os
import pykube
import pytest
from rx import subject as rx_subject
from dependency_injector import providers
# from krules_core.base_functions import Filter, Process

from krules_core.core import RuleFactory
from krules_core import RuleConst
from krules_core.providers import event_router_factory, subject_storage_factory, proc_events_rx_factory
from krules_core.tests.subject.sqlite_storage import SQLLiteSubjectStorage

from .functions import (
    K8sObjectCreate,
    K8sObjectUpdate,
    K8sObjectsQuery,
    K8sObjectDelete,
)


def setup_module(_):
    extra_kwargs = {}
    if os.environ.get("API_URL", False):
        extra_kwargs.update({
            "override_api_url": os.environ.get("API_URL")
        })

    subject_storage_factory.override(
        providers.Factory(lambda *args, **kwargs: SQLLiteSubjectStorage(args[0], ":memory:"))
    )

    proc_events_rx_factory.override(providers.Singleton(rx_subject.ReplaySubject))


def teardown_module(_):
    subject_storage_factory.reset_last_overriding()


@pytest.fixture
def api():
    if os.environ.get("API_URL", False):
        config = pykube.KubeConfig.from_url(os.environ.get("API_URL"))
    else:
        config = pykube.KubeConfig.from_env()
    return pykube.HTTPClient(config)


@pytest.fixture
def namespace(api):
    if os.environ.get("NAMESPACE", False):
        return os.environ.get("NAMESPACE")
    return api.config.namespace


def _assert(expr, msg="test failed"):
    print("################ ", expr)
    assert expr, msg
    return True


def test_create(api, namespace):
    RuleFactory.create(
        'test-k8s-create',
        subscribe_to="test-create",
        data={
            RuleConst.PROCESSING: [
                K8sObjectCreate({
                    "apiVersion": "v1",
                    "kind": "Pod",
                    "metadata": {
                        "name": "test-pod-1",
                        "labels": {
                            "app": "pytest-temp"
                        }
                    },
                    "spec": {
                        "containers": [{
                            "name": "hello",
                            "image": "karthequian/helloworld"
                        }]
                    }
                }),
                K8sObjectCreate({
                    "apiVersion": "v1",
                    "kind": "Pod",
                    "metadata": {
                        "name": "test-pod-2",
                        "labels": {
                            "app": "pytest-temp"
                        }
                    },
                    "spec": {
                        "containers": [{
                            "name": "hello",
                            "image": "karthequian/helloworld"
                        }]
                    }
                })
            ]
        }
    )

    proc_events_rx_factory().subscribe(
        lambda x: x[RuleConst.RULENAME] == 'test-k8s-create' and
                  _assert(
                      not x[RuleConst.GOT_ERRORS] and x[RuleConst.PASSED],
                      "test-k8s-create proc failed"
                  )
    )

    event_router_factory().route("test-create", "some", {})

    pykube.Pod.objects(api).filter(namespace=namespace)
    objs = pykube.Pod.objects(api).filter(namespace=namespace, selector={"app": "pytest-temp"})
    # wait for pods ready
    watch = objs.watch()
    ready = 0
    for ev in watch:
        print(ev.type, ev.object.name, ev.object.ready)
        if ev.object.ready:
            ready += 1
        if ready == 2:
            break
    assert len(objs) == 2

    event_router_factory().unregister_all()


def test_update(api, namespace):
    RuleFactory.create(
        'test-k8s-update-in-context',
        subscribe_to="test-update-in-context",
        data={
            RuleConst.PROCESSING: [
                # update from subject
                K8sObjectUpdate(
                    patch=lambda obj: (
                        obj["metadata"]["labels"].update({
                            "updated": "pod-1"
                        })
                    ),
                    name="test-pod-1",
                    kind="Pod",
                    apiversion="v1",
                    namespace=namespace
                ),
            ]
        }
    )

    proc_events_rx_factory().subscribe(
        lambda x: x[RuleConst.RULENAME] == 'test-k8s-update-in-context' and
                  _assert(
                      x[RuleConst.PASSED] and x[RuleConst.GOT_ERRORS] is False,
                      "test-k8s-update-in-context proc failed"
                  )
    )

    objs = pykube.Pod.objects(api).filter(namespace=namespace, selector={"app": "pytest-temp"})
    watch = objs.watch().filter(field_selector={"metadata.name": "test-pod-1"})
    event_router_factory().route(
        "test-update-in-context",
        f"k8s:/api/v1/namespaces/{namespace}/pods/test-pod-1", {}
    )

    try:
        for ev in watch:
            if ev.type == 'ADDED' and ev.object.obj["metadata"]["labels"]["updated"] == "pod-1":
                break
    except:
        assert False, "update pod-1 event failed"

    RuleFactory.create(
        'test-k8s-update-off-context',
        subscribe_to="test-update-off-context",
        data={
            RuleConst.PROCESSING: [
                # update from subject
                K8sObjectUpdate(
                    lambda obj: (
                        obj["metadata"]["labels"].update({
                            "updated": "pod-2"
                        })
                    ),
                    name="test-pod-2",
                    apiversion="v1",
                    kind="Pod",
                    namespace=namespace
                ),
            ]
        }
    )

    proc_events_rx_factory().subscribe(
        lambda x: x[RuleConst.RULENAME] == 'test-k8s-update-off-context' and
                  _assert(
                      x[RuleConst.GOT_ERRORS] is False and x[RuleConst.PASSED],
                      "test-k8s-update-off-context proc failed"
                  )
    )

    objs = pykube.Pod.objects(api).filter(namespace=namespace, selector={"app": "pytest-temp"})
    watch = objs.watch().filter(field_selector={"metadata.name": "test-pod-2"})
    event_router_factory().route(
        "test-update-off-context",
        "none", {}
    )

    try:
        for ev in watch:
            if ev.type == 'ADDED' and ev.object.obj["metadata"]["labels"]["updated"] == "pod-2":
                break
    except:
        assert False, "update pod-2 event failed"

    event_router_factory().unregister_all()


def tests_query_foreach(api, namespace):
    RuleFactory.create(
        'test-k8s-query-foreach',
        subscribe_to="test-query-foreach",
        data={
            RuleConst.PROCESSING: [
                K8sObjectsQuery(
                    foreach=lambda obj: (
                        obj.obj["metadata"]["labels"].update({
                            "updated": "foreach"
                        }),
                        obj.update()
                    ),
                    apiversion="v1",
                    kind="Pod",
                    namespace=namespace,
                    selector={
                        "app": "pytest-temp"
                    }
                ),
            ]
        }
    )

    proc_events_rx_factory().subscribe(
        lambda x: x[RuleConst.RULENAME] == 'test-k8s-query-foreach' and
                  _assert(
                      x[RuleConst.GOT_ERRORS] is False and x[RuleConst.PASSED],
                      "test-k8s-query-foreach proc failed"
                  )
    )

    watch = pykube.Pod.objects(api).filter(namespace=namespace, selector={"app": "pytest-temp"}).watch()
    event_router_factory().route(
        "test-query-foreach",
        "none", {}
    )

    try:
        updated = 0
        for ev in watch:
            if ev.type == 'ADDED' and ev.object.obj["metadata"]["labels"]["updated"] == "foreach":
                updated += 1
            if updated == 2:
                break
    except:
        assert False, "query update event failed"

    assert updated == 2

    event_router_factory().unregister_all()


# NOTE: delete tests K8SObjectsQuery#returns indirectly
def test_delete(api, namespace):
    RuleFactory.create(
        'test-k8s-delete-in-context',
        subscribe_to="test-delete-in-context",
        data={
            RuleConst.PROCESSING: [
                K8sObjectDelete(
                    name="test-pod-1",
                    apiversion="v1",
                    kind="Pod",
                    namespace=namespace)
            ]
        }
    )

    objs = pykube.Pod.objects(api).filter(namespace=namespace, selector={"app": "pytest-temp"})
    watch = objs.watch().filter(field_selector={"metadata.name": "test-pod-1"})
    proc_events_rx_factory().subscribe(
        lambda x: x[RuleConst.RULENAME] == 'test-k8s-delete-in-context' and
                  _assert(
                      x[RuleConst.GOT_ERRORS] is False and x[RuleConst.PASSED],
                      "test-k8s-delete-in-context proc failed"
                  )
    )
    event_router_factory().route(
        "test-delete-in-context",
        f"k8s:/api/v1/namespaces/{namespace}/pods/test-pod-1", {}
    )

    for ev in watch:
        if ev.type == "DELETED":
            break
    assert len(objs) == 1

    RuleFactory.create(
        'test-k8s-delete-off-context',
        subscribe_to="test-delete-off-context",
        data={
            RuleConst.PROCESSING: [
                # update from subject
                K8sObjectDelete(
                    name="test-pod-2",
                    apiversion="v1",
                    kind="Pod",
                    namespace=namespace
                )
            ]
        }
    )

    objs = pykube.Pod.objects(api).filter(namespace=namespace, selector={"app": "pytest-temp"})
    watch = objs.watch().filter(field_selector={"metadata.name": "test-pod-2"})
    proc_events_rx_factory().subscribe(
        lambda x: x[RuleConst.RULENAME] == 'test-k8s-delete-off-context' and
                  _assert(
                      x[RuleConst.GOT_ERRORS] is False and x[RuleConst.PASSED],
                      "test-k8s-delete-off-context proc failed"
                  )
    )
    event_router_factory().route(
        "test-delete-off-context",
        f"k8s:/api/v1/namespaces/{namespace}", {}
    )

    for ev in watch:
        if ev.type == "DELETED":
            break
    assert len(objs) == 0

    event_router_factory().unregister_all()



# Copyright 2019 The KRules Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime, timezone
import inspect
import uuid

import pykube
from krules_core.base_functions import RuleFunctionBase

from krules_core.providers import subject_factory


def k8s_event_create(api, producer, action, message, reason, type,
                     reporting_component=None, reporting_instance=None, involved_object=None, namespace=None,
                     source_component=None, first_timestamp=None, last_timestamp=None):
    """
    *Crate a Kubernetes reosurce of type v1/Event.*

    """
    dt_now = datetime.now(timezone.utc).astimezone().isoformat()

    if first_timestamp is None:
        first_timestamp = dt_now
    if last_timestamp is None:
        last_timestamp = dt_now

    obj = {
        "apiVersion": "v1",
        "kind": "Event",
        "eventTime": datetime.now(timezone.utc).astimezone().isoformat(),
        "firstTimestamp": first_timestamp,
        "lastTimestamp": last_timestamp,
        "metadata": {
            "name": "{}.{}".format(producer, uuid.uuid4().hex[:16])
        },
        "action": action,
        "message": message,
        "reason": reason,
        # "source": {
        #     "component": source_component,
        # },
        "type": type
    }

    if reporting_component:
        obj.update({
            "reportingComponent": reporting_component
        })
    if reporting_instance:
        obj.update({
            "reportingInstance": reporting_instance
        })
    if source_component:
        obj.update({
            "source": {
                "component": source_component
            }
        })

    namespace = namespace is not None and namespace \
                or involved_object is not None and involved_object.get("metadata", {}).get("namespace") \
                or None
    if namespace:
        obj["metadata"].update({
            "namespace": namespace
        })

    if involved_object:
        obj.update({
            "involvedObject": {
                "apiVersion": involved_object.get("apiVersion"),
                "kind": involved_object.get("kind"),
                "name": involved_object.get("metadata").get("name"),
                "namespace": involved_object.get("metadata").get("namespace"),
                "resourceVersion": involved_object.get("metadata").get("resourceVersion"),
                "uid": involved_object.get("metadata").get("uid")
            },
        })

    pykube.Event(api, obj).create()


class K8sRuleFunctionBase(RuleFunctionBase):

    def _get_object(self, apiversion, kind):

        api = self.payload.get("_k8s_api_client")
        if api is None:
            config = pykube.KubeConfig.from_env()
            api = pykube.HTTPClient(config)
            self.payload["_k8s_api_client"] = api

        if apiversion is None and "apiversion" in self.subject.get_ext_props():
            apiversion = self.subject.get_ext("apiversion")

        if kind is None and "kind" in self.subject.get_ext_props():
            kind = self.subject.get_ext("kind")

        return pykube.object_factory(api, apiversion, kind)


class K8sObjectsQuery(K8sRuleFunctionBase):
    """
    *Search for a list of K8s resources using the given filters.*

    ::

        rulesdata = [
            {
                rulename: "...",
                subscibre_to: "...",
                ruledata: {
                    filters: [
                        Filter(
                            K8sObjectsQuery(
                                kind="Pod",
                                returns=lambda payload: lambda obj: (
                                    payload.set_default("pods", obj)
                                )
                            )
                        )
                    ]
                    processing: [
                        ...
                    ]
                }
            }
        ]
    """

    def execute(self, apiversion=None, kind=None, foreach=None, returns=None, **filters):

        """
        Args:

            apiversion: resources api version, if None get this value from subject extended properties[default None]
            kind: resources kind version, if None get this value from subject extended properties[default None]
            foreach: an optional callable that will be execute for each query element[default None]
            returns: an optional callable that will received as parameters the whole query elements[default None]
            **filters: query kwargs
        """

        obj = self._get_object(apiversion, kind)

        qobjs = obj.objects(self.payload.get("_k8s_api_client")).filter(**filters)

        if foreach is not None:
            for obj in qobjs:
                while True:
                    try:
                        foreach(obj)
                        break
                    except pykube.exceptions.HTTPError as ex:
                        if ex.code == 409:
                            continue
                        else:
                            raise ex

        if returns is not None:
            while True:
                try:
                    return returns(qobjs)
                except pykube.exceptions.HTTPError as ex:
                    if ex.code == 409:
                        continue
                    else:
                        raise ex
        return len(qobjs)


class K8sObjectUpdate(K8sRuleFunctionBase):

    """
    *Update a K8s object.*

    ::

        rulesdata = [
            {
                rulename: "...",
                subscibre_to: "...",
                ruledata: {
                    filters: [
                        ...
                    ]
                    processing: [
                        K8sObjectUpdate(
                            lambda obj: (
                                obj["metadata"]["labels"].update({
                                    "updated": "my-pod"
                                })
                            ),
                            name="my-pod",
                            apiversion="v1",
                            kind="Pod",
                            namespace="default"
                        ),
                    ]
                }
            }
        ]
    """

    def execute(self, patch, name, apiversion, kind, subresource=None, is_strategic=True, **filters):

        """
        Args:
            patch: could be a callable that receive a dict or a dict itself
            name: resource name, if None get this value from subject extended properties[default None]
            apiversion: resource api version, if None get this value from subject extended properties[default None]
            kind: resource kind version, if None get this value from subject extended properties[default None]
            subresource: subresource parameter of pykube object update[default None]
            is_strategic: is_strategic parameter of pykube object update[default True]
            **filters: query kwargs
        """

        obj = self._get_object(apiversion, kind)
        obj = obj.objects(self.payload.get("_k8s_api_client")).filter(**filters).get(name=name)

        # retrocompatibility
        if inspect.isfunction(patch):
            patch(obj.obj)
        else:
            obj.obj.update(patch)

        while True:
            try:
                obj.update(subresource=subresource, is_strategic=is_strategic)
                break
            except pykube.exceptions.HTTPError as ex:
                if ex.code == 409:
                    continue
                else:
                    raise ex


# class K8sObjectPatch(K8sRuleFunctionBase):
#
#     def execute(self, patch, name=None, apiversion=None, kind=None, subresource=None, **filters):
#
#         if name is None:
#             name = self.subject.get_ext("name")
#
#         obj = self._get_object(apiversion, kind)
#
#         # we are implicitly referring to the resource in the subject
#         if kind is None and apiversion is None \
#                 and "namespace" not in filters \
#                 and "namespace" in self.subject.get_ext_props() \
#                 and self.subject.ext_namespace is not None:
#             filters.update({
#                 "namespace": self.subject.get_ext("namespace")
#             })
#
#         obj = obj.objects(self.payload.get("_k8s_api_client")).filter(**filters).get(name=name)
#
#         while True:
#             try:
#                 obj.patch(patch, subresource=subresource)
#                 break
#             except pykube.exceptions.HTTPError as ex:
#                 if ex.code == 409:
#                     continue
#                 else:
#                     raise ex


class K8sObjectCreate(K8sRuleFunctionBase):
    """
    *Create a K8s object.*

    ::

        rulesdata = [
            {
                rulename: "...",
                subscibre_to: "...",
                ruledata: {
                    filters: [
                        ...
                    ]
                    processing: [
                        K8sObjectCreate({
                            "apiVersion": "v1",
                            "kind": "Pod",
                            "metadata": {
                                "name": "my-pod",
                                "labels": {
                                    "app": "pytest-temp"
                                }
                            },
                            "spec": {
                                "containers": [{
                                    "name": "hello",
                                    "image": "karthequian/helloworld"
                                }]
                            }
                        })
                    ]
                }
            }
        ]
    """

    def execute(self, obj):
        """
        Args:
            obj: dict representing the structure of the object you want to create.
        """

        apiversion = obj.get("apiVersion")
        kind = obj.get("kind")
        while True:
            try:
                obj_kind = self._get_object(apiversion, kind)
                api = self.payload.get("_k8s_api_client")
                obj_ref = obj_kind(api, obj)
                if not obj_ref.exists():
                    obj_ref.create()
                break
            except pykube.exceptions.HTTPError as ex:
                if ex.code == 409:
                    continue
                else:
                    raise ex


class K8sObjectDelete(K8sObjectsQuery):
    """
    *Delete a K8s object.*

    ::

        rulesdata = [
            {
                rulename: "...",
                subscibre_to: "...",
                ruledata: {
                    filters: [
                        ...
                    ]
                    processing: [
                        K8sObjectCreate(
                            name="my-pod",
                            apiversion="v1",
                            kind="Pod",
                            namespace="default"
                        )
                    ]
                }
            }
        ]
    """

    def execute(self, name, apiversion=None, kind=None, **filters):
        """
        Args:
            name: resource name, if None get this value from subject extended properties[default None]
            apiversion: resource api version, if None get this value from subject extended properties[default None]
            kind: resource kind version, if None get this value from subject extended properties[default None]
            **filters: query kwargs
        """

        super().execute(apiversion=apiversion, kind=kind, returns=lambda qobjs: (
            qobjs.get(name=name).delete()
        ), **filters)



# Copyright 2019 The KRules Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from setuptools import setup, find_packages


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name='krules-k8s-functions',
    version="0.11",
    author="Alberto Degli Esposti",
    author_email="alberto@arispot.tech",
    description="KRules k8s subjects storage implementation",
    license="Apache Licence 2.0",
    keywords="krules kubernetes k8s python python knative",
    url="https://github.com/airspot-dev/krules-subjects-storage-k8s.git",
    packages=find_packages(),
    long_description=read('README.md'),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: Apache Software License",
    ],
    install_requires=[
        'krules-core==0.11',
        'pykube-ng==21.3.0',
    ],
    setup_requires=[
        'pytest-runner',
    ],
    tests_require=[
        'pytest',
    ],
)